@@ -15,13 +15,34 @@ module ddx_cinterface
         type(ddx_params_type)    :: params
         type(ddx_constants_type) :: constants
         type(ddx_workspace_type) :: workspace
-        character(len=255)       :: error_message  ! most recent error message (if any)
     end type ddx_setup
 
 contains
 !
 ! Generic stuff
 !
+subroutine ddx_get_banner(message, maxlen) bind(C)
+    integer(c_int), intent(in), value :: maxlen
+    character(len=1, kind=C_char), intent(out) :: message(maxlen)
+    integer :: length, i
+    character :: ch
+    character (len=4095) :: header
+    call get_banner(header)
+    message(maxlen) = c_null_char
+    length = min(maxlen-1, 4095)
+    do i = length, 1, -1
+        if (header(i:i) .eq. ' ') then
+            length = i-1
+        else
+            exit
+        endif
+    enddo
+    message(length + 1) = c_null_char
+    do i = 1, length
+        message(i) = header(i:i)
+    enddo
+end
+
 function ddx_supported_lebedev_grids(n, grids) result(c_ngrids) bind(C)
     integer(c_int), intent(in), value ::  n
     integer(c_int), intent(out) :: grids(n)
@@ -64,76 +85,52 @@ subroutine ddx_scaled_ylm(c_ddx, lmax, x, sphere, ylm) bind(C)
     end associate
 end
 
-subroutine ddx_nuclear_contributions(c_ddx, nsph, ncav, nbasis, phi, gradphi, psi) bind(C)
-    type(c_ptr), intent(in), value :: c_ddx
-    integer(c_int), intent(in), value :: nsph, ncav, nbasis
-    type(ddx_setup), pointer :: ddx
-    real(c_double), intent(out) :: phi(ncav)
-    real(c_double), intent(out) :: gradphi(3, ncav)
-    real(c_double), intent(out) :: psi(nbasis, nsph)
-    real(dp) :: hessianphi(3, 3, ncav)  ! not used, hence not allocated.
-    call c_f_pointer(c_ddx, ddx)
-    call mkrhs(ddx%params, ddx%constants, ddx%workspace, 1, phi, 1, gradphi, 0, &
-            & hessianphi, psi)
-end
-
 !
 ! Setup object
 !
-function ddx_allocate_model(model, enable_force, solvent_epsilon, solvent_kappa, eta, lmax, &
+function ddx_allocate_model(model, enable_force, solvent_epsilon, solvent_kappa, eta, se, lmax, &
         & n_lebedev, incore, maxiter, jacobi_n_diis, enable_fmm, fmm_multipole_lmax, fmm_local_lmax, &
         & n_proc, n_spheres, sphere_charges, sphere_centres, &
-        !& sphere_radii, printfctn, info) result(c_ddx) bind(C)
-        & sphere_radii, info) result(c_ddx) bind(C)
+        & sphere_radii, length_logfile, c_logfile) result(c_ddx) bind(C)
     integer(c_int), intent(in), value :: model, enable_force, lmax, n_lebedev, maxiter, &
-        & incore, jacobi_n_diis, enable_fmm, fmm_multipole_lmax, fmm_local_lmax, n_proc, n_spheres
+        & incore, jacobi_n_diis, enable_fmm, fmm_multipole_lmax, fmm_local_lmax, n_proc, &
+        & n_spheres, length_logfile
     real(c_double), intent(in) :: sphere_charges(n_spheres), sphere_centres(3, n_spheres), &
         & sphere_radii(n_spheres)
-    real(c_double), intent(in), value :: eta, solvent_epsilon, solvent_kappa
-    integer(c_int), intent(out) :: info
+    real(c_double), intent(in), value :: eta, se, solvent_epsilon, solvent_kappa
     !type(c_funptr), value :: printfctn
     type(c_ptr) :: c_ddx
-    integer :: passproc, gmresr_j, gmresr_dim, itersolver
-    real(dp) :: se
+    integer :: passproc, i
     type(ddx_setup), pointer :: ddx
-
-    ! interface
-    !     subroutine print_characters(string) bind(C)
-    !         character(kind=c_char), dimension(*), intent(in) :: string
-    !     end subroutine
-    ! end interface
-    ! procedure(print_characters), pointer :: cprint
-    ! call c_f_procpointer(printfctn, cprint)  ! Convert C to fortran procedure pointer
+    character(len=1, kind=C_char), intent(in) :: c_logfile(length_logfile)
+    character(len=255) :: logfile
 
     ! Allocate DDX object
     allocate(ddx)
     c_ddx = c_loc(ddx)
 
+    ! Convert to Fortran objects
     passproc = n_proc
-    se = 0.0        ! Hard-code centred regularisation
-    itersolver = 1  ! Hard-code Jacobi DIIS
-    gmresr_j   = 1  ! Dummy as GMRES not used
-    gmresr_dim = 5  ! Dummy as GMRES not used
+    logfile = ''
+    do i = 1, min(length_logfile, 255)
+        if (c_logfile(i) .ne. c_null_char) then
+            logfile(i:i) = c_logfile(i)
+        endif
+    enddo
 
-    ddx%error_message = "No error"
     call params_init(model, enable_force, solvent_epsilon, solvent_kappa, eta, se, lmax, &
-        & n_lebedev, incore, itersolver, maxiter, jacobi_n_diis, gmresr_j, gmresr_dim, enable_fmm, &
+        & n_lebedev, incore, maxiter, jacobi_n_diis, enable_fmm, &
         & fmm_multipole_lmax, fmm_local_lmax, passproc, n_spheres, sphere_charges, &
-        & sphere_centres, sphere_radii, print_func_default, ddx%params, info)
-    if (info .ne. 0) then
-        ddx%error_message = ddx%params%error_message
+        & sphere_centres, sphere_radii, logfile, ddx%params)
+    if (ddx%params%error_flag .ne. 0) then
         return
     endif
-    info = 0
-    call constants_init(ddx%params, ddx%constants, info)
-    if (info .ne. 0) then
-        ddx%error_message = ddx%constants%error_message
+    call constants_init(ddx%params, ddx%constants)
+    if (ddx%constants%error_flag .ne. 0) then
         return
     endif
-    info = 0
-    call workspace_init(ddx%params, ddx%constants, ddx%workspace, info)
-    if (info .ne. 0) then
-        ddx%error_message = ddx%workspace%error_message
+    call workspace_init(ddx%params, ddx%constants, ddx%workspace)
+    if (ddx%workspace%error_flag .ne. 0) then
         return
     endif
 end function
@@ -141,26 +138,53 @@ end function
 subroutine ddx_deallocate_model(c_ddx) bind(C)
     type(c_ptr), intent(in), value :: c_ddx
     type(ddx_setup), pointer :: ddx
-    integer :: info
     call c_f_pointer(c_ddx, ddx)
-    call params_deinit(ddx%params, info)
-    call constants_free(ddx%constants, info)
-    call workspace_free(ddx%workspace, info)
+    call params_deinit(ddx%params)
+    call constants_free(ddx%constants)
+    call workspace_free(ddx%workspace)
     deallocate(ddx)
 end
 
+function ddx_get_error_flag(c_ddx) result(has_error) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx
+    type(ddx_setup), pointer :: ddx
+    integer(c_int) :: has_error
+    call c_f_pointer(c_ddx, ddx)
+    has_error = 0
+    if (ddx%params%error_flag .ne. 0) then
+        has_error = ddx%params%error_flag
+    elseif (ddx%constants%error_flag .ne. 0) then
+        has_error = ddx%constants%error_flag
+    elseif (ddx%workspace%error_flag .ne. 0) then
+        has_error = ddx%workspace%error_flag
+    endif
+end
+
 subroutine ddx_get_error_message(c_ddx, message, maxlen) bind(C)
     type(c_ptr), intent(in), value :: c_ddx
     integer(c_int), intent(in), value :: maxlen
     character(len=1, kind=C_char), intent(out) :: message(maxlen)
+    character(len=255) :: error_message
     type(ddx_setup), pointer :: ddx
     integer :: length, i
     character :: ch
     call c_f_pointer(c_ddx, ddx)
+    ! Get the actual error message from the collection of structs
+    if (ddx%params%error_flag .ne. 0) then
+        error_message = ddx%params%error_message
+    elseif (ddx%constants%error_flag .ne. 0) then
+        error_message = ddx%constants%error_message
+    elseif (ddx%workspace%error_flag .ne. 0) then
+        error_message = ddx%workspace%error_message
+    else
+        error_message = ''
+    endif
+
+    ! Convert to C message
     message(maxlen) = c_null_char
     length = min(maxlen-1, 255)
     do i = length, 1, -1
-        if (ddx%error_message(i:i) .eq. ' ') then
+        if (error_message(i:i) .eq. ' ') then
             length = i-1
         else
             exit
@@ -168,7 +192,32 @@ subroutine ddx_get_error_message(c_ddx, message, maxlen) bind(C)
     enddo
     message(length + 1) = c_null_char
     do i = 1, length
-        message(i) = ddx%error_message(i:i)
+        message(i) = error_message(i:i)
+    enddo
+end
+
+
+subroutine ddx_get_logfile(c_ddx, message, maxlen) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx
+    integer(c_int), intent(in), value :: maxlen
+    character(len=1, kind=C_char), intent(out) :: message(maxlen)
+    type(ddx_setup), pointer :: ddx
+    integer :: length, i
+    character :: ch
+    call c_f_pointer(c_ddx, ddx)
+    ! Convert to C message
+    message(maxlen) = c_null_char
+    length = min(maxlen-1, 255)
+    do i = length, 1, -1
+        if (ddx%params%output_filename(i:i) .eq. ' ') then
+            length = i-1
+        else
+            exit
+        endif
+    enddo
+    message(length + 1) = c_null_char
+    do i = 1, length
+        message(i) = ddx%params%output_filename(i:i)
     enddo
 end
 
@@ -293,6 +342,14 @@ function ddx_get_solvent_kappa(c_ddx) result(c_kappa) bind(C)
     c_kappa = ddx % params % kappa
 end function
 
+function ddx_get_shift(c_ddx) result(c_se) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx
+    type(ddx_setup), pointer :: ddx
+    real(c_double) :: c_se
+    call c_f_pointer(c_ddx, ddx)
+    c_se = ddx % params % se
+end function
+
 subroutine ddx_get_sphere_charges(c_ddx, nsph, c_charge) bind(C)
     type(c_ptr), intent(in), value :: c_ddx
     integer(c_int), intent(in), value :: nsph
@@ -417,161 +474,206 @@ end subroutine
 !
 ! Cosmo
 !
-subroutine ddx_cosmo_fill_guess(c_ddx, c_state) bind(C)
+! Setup the problem in the state
+subroutine ddx_cosmo_setup(c_ddx, c_state, ncav, nbasis, nsph, phi_cav, psi) bind(C)
     type(c_ptr), intent(in), value :: c_ddx, c_state
+    integer(c_int), intent(in), value :: ncav, nbasis, nsph
+    real(c_double), intent(in) :: phi_cav(ncav), psi(nbasis, nsph)
     type(ddx_setup), pointer :: ddx
     type(ddx_state_type), pointer :: state
     call c_f_pointer(c_ddx, ddx)
     call c_f_pointer(c_state, state)
-    call ddcosmo_guess(ddx%params, ddx%constants, state)
-end
-
-subroutine ddx_cosmo_solve(c_ddx, c_state, ncav, phi, tol) bind(C)
-    type(c_ptr), intent(in), value :: c_ddx, c_state
-    type(ddx_setup), pointer :: ddx
-    type(ddx_state_type), pointer :: state
-    integer(c_int), intent(in), value :: ncav
-    real(c_double), intent(in), value :: tol
-    real(c_double), intent(in) :: phi(ncav)
-    call c_f_pointer(c_ddx, ddx)
-    call c_f_pointer(c_state, state)
-    call ddcosmo_solve(ddx%params, ddx%constants, ddx%workspace, state, phi, tol)
+    call ddcosmo_setup(ddx%params, ddx%constants, ddx%workspace, state, phi_cav, psi)
 end subroutine
 
-subroutine ddx_cosmo_adjoint(c_ddx, c_state, nbasis, nsph, psi, tol) bind(C)
+! Put a guess for the problem into the state (optional)
+subroutine ddx_cosmo_guess(c_ddx, c_state) bind(C)
     type(c_ptr), intent(in), value :: c_ddx, c_state
     type(ddx_setup), pointer :: ddx
     type(ddx_state_type), pointer :: state
-    integer(c_int), intent(in), value :: nbasis, nsph
-    real(c_double), intent(in), value :: tol
-    real(c_double), intent(in) :: psi(nbasis, nsph)
     call c_f_pointer(c_ddx, ddx)
     call c_f_pointer(c_state, state)
-    call ddcosmo_adjoint(ddx%params, ddx%constants, ddx%workspace, state, psi, tol)
+    call ddcosmo_guess(ddx%params, ddx%constants, ddx%workspace, state)
 end
 
-subroutine ddx_cosmo_forces(c_ddx, c_state, nbasis, nsph, ncav, phi, gradphi, psi, forces) bind(C)
+! Solve the problem
+subroutine ddx_cosmo_solve(c_ddx, c_state, tol) bind(C)
     type(c_ptr), intent(in), value :: c_ddx, c_state
     type(ddx_setup), pointer :: ddx
     type(ddx_state_type), pointer :: state
-    integer(c_int), intent(in), value :: ncav, nbasis, nsph
-    real(c_double), intent(in) :: phi(ncav), gradphi(3, ncav), psi(nbasis, nsph)
+    real(c_double), intent(in), value :: tol
+    call c_f_pointer(c_ddx, ddx)
+    call c_f_pointer(c_state, state)
+    call ddcosmo_solve(ddx%params, ddx%constants, ddx%workspace, state, tol)
+end subroutine
+
+! Put a guess for the adjoint problem into the state (optional)
+subroutine ddx_cosmo_guess_adjoint(c_ddx, c_state) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx, c_state
+    type(ddx_setup), pointer :: ddx
+    type(ddx_state_type), pointer :: state
+    call c_f_pointer(c_ddx, ddx)
+    call c_f_pointer(c_state, state)
+    call ddcosmo_guess_adjoint(ddx%params, ddx%constants, ddx%workspace, state)
+end
+
+! Solve the adjoint problem inside the state
+subroutine ddx_cosmo_solve_adjoint(c_ddx, c_state, tol) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx, c_state
+    type(ddx_setup), pointer :: ddx
+    type(ddx_state_type), pointer :: state
+    real(c_double), intent(in), value :: tol
+    call c_f_pointer(c_ddx, ddx)
+    call c_f_pointer(c_state, state)
+    call ddcosmo_solve_adjoint(ddx%params, ddx%constants, ddx%workspace, state, tol)
+end
+
+! Compute the forces
+subroutine ddx_cosmo_solvation_force_terms(c_ddx, c_state, nsph, forces) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx, c_state
+    type(ddx_setup), pointer :: ddx
+    type(ddx_state_type), pointer :: state
+    integer(c_int), intent(in), value :: nsph
     real(c_double), intent(out) :: forces(3, nsph)
     call c_f_pointer(c_ddx, ddx)
     call c_f_pointer(c_state, state)
-    call ddcosmo_forces(ddx%params, ddx%constants, ddx%workspace, state, phi, gradphi, psi, forces)
+    call ddcosmo_solvation_force_terms(ddx%params, ddx%constants, ddx%workspace, state, forces)
 end
 
 !
 ! PCM
 !
-subroutine ddx_pcm_fill_guess(c_ddx, c_state) bind(C)
+subroutine ddx_pcm_setup(c_ddx, c_state, ncav, nbasis, nsph, phi_cav, psi) bind(C)
     type(c_ptr), intent(in), value :: c_ddx, c_state
+    integer(c_int), intent(in), value :: ncav, nbasis, nsph
+    real(c_double), intent(in) :: phi_cav(ncav), psi(nbasis, nsph)
     type(ddx_setup), pointer :: ddx
     type(ddx_state_type), pointer :: state
     call c_f_pointer(c_ddx, ddx)
     call c_f_pointer(c_state, state)
-    call ddpcm_guess(ddx%params, ddx%constants, state)
-end
-
-subroutine ddx_pcm_solve(c_ddx, c_state, ncav, phi, tol) bind(C)
-    type(c_ptr), intent(in), value :: c_ddx, c_state
-    type(ddx_setup), pointer :: ddx
-    type(ddx_state_type), pointer :: state
-    integer(c_int), intent(in), value :: ncav
-    real(c_double), intent(in), value :: tol
-    real(c_double), intent(in) :: phi(ncav)
-    call c_f_pointer(c_ddx, ddx)
-    call c_f_pointer(c_state, state)
-    call ddpcm_solve(ddx%params, ddx%constants, ddx%workspace, state, phi, tol)
+    call ddpcm_setup(ddx%params, ddx%constants, ddx%workspace, state, phi_cav, psi)
 end subroutine
 
-subroutine ddx_pcm_adjoint(c_ddx, c_state, nbasis, nsph, psi, tol) bind(C)
+subroutine ddx_pcm_guess(c_ddx, c_state) bind(C)
     type(c_ptr), intent(in), value :: c_ddx, c_state
     type(ddx_setup), pointer :: ddx
     type(ddx_state_type), pointer :: state
-    integer(c_int), intent(in), value :: nbasis, nsph
-    real(c_double), intent(in), value :: tol
-    real(c_double), intent(in) :: psi(nbasis, nsph)
     call c_f_pointer(c_ddx, ddx)
     call c_f_pointer(c_state, state)
-    call ddpcm_adjoint(ddx%params, ddx%constants, ddx%workspace, state, psi, tol)
+    call ddpcm_guess(ddx%params, ddx%constants, ddx%workspace, state)
 end
 
-subroutine ddx_pcm_forces(c_ddx, c_state, nbasis, nsph, ncav, phi, gradphi, psi, forces) bind(C)
+subroutine ddx_pcm_solve(c_ddx, c_state, tol) bind(C)
     type(c_ptr), intent(in), value :: c_ddx, c_state
     type(ddx_setup), pointer :: ddx
     type(ddx_state_type), pointer :: state
-    integer(c_int), intent(in), value :: ncav, nbasis, nsph
-    real(c_double), intent(in) :: phi(ncav), gradphi(3, ncav), psi(nbasis, nsph)
+    real(c_double), intent(in), value :: tol
+    call c_f_pointer(c_ddx, ddx)
+    call c_f_pointer(c_state, state)
+    call ddpcm_solve(ddx%params, ddx%constants, ddx%workspace, state, tol)
+end subroutine
+
+subroutine ddx_pcm_guess_adjoint(c_ddx, c_state) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx, c_state
+    type(ddx_setup), pointer :: ddx
+    type(ddx_state_type), pointer :: state
+    call c_f_pointer(c_ddx, ddx)
+    call c_f_pointer(c_state, state)
+    call ddpcm_guess_adjoint(ddx%params, ddx%constants, ddx%workspace, state)
+end
+
+subroutine ddx_pcm_solve_adjoint(c_ddx, c_state, tol) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx, c_state
+    type(ddx_setup), pointer :: ddx
+    type(ddx_state_type), pointer :: state
+    real(c_double), intent(in), value :: tol
+    call c_f_pointer(c_ddx, ddx)
+    call c_f_pointer(c_state, state)
+    call ddpcm_solve_adjoint(ddx%params, ddx%constants, ddx%workspace, state, tol)
+end
+
+subroutine ddx_pcm_solvation_force_terms(c_ddx, c_state, nsph, forces) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx, c_state
+    type(ddx_setup), pointer :: ddx
+    type(ddx_state_type), pointer :: state
+    integer(c_int), intent(in), value :: nsph
     real(c_double), intent(out) :: forces(3, nsph)
     call c_f_pointer(c_ddx, ddx)
     call c_f_pointer(c_state, state)
-    call ddpcm_forces(ddx%params, ddx%constants, ddx%workspace, state, phi, gradphi, psi, forces)
+    call ddpcm_solvation_force_terms(ddx%params, ddx%constants, ddx%workspace, state, forces)
 end
 
 !
 ! multipolar solutes
 !
-
-subroutine ddx_build_g(c_ddx, multipoles, mmax, phi_cav, e_cav, g_cav, &
-        & nsph, ncav) bind(C)
+subroutine ddx_multipole_electrostatics_0(c_ddx, nsph, ncav, nmultipoles, multipoles, &
+        & phi_cav) bind(C)
     type(c_ptr), intent(in), value :: c_ddx
     type(ddx_setup), pointer :: ddx
-    integer(c_int), intent(in), value :: mmax, nsph, ncav
-    real(c_double), intent(in) :: multipoles((mmax + 1)**2, nsph)
-    real(c_double), intent(out) :: phi_cav(ncav), e_cav(3, ncav), &
-        & g_cav(3, 3, ncav)
-    call c_f_pointer(c_ddx, ddx)
-    call build_g(ddx%params, ddx%constants, ddx%workspace, multipoles, mmax, &
-        & phi_cav, e_cav, g_cav)
-end
-
-subroutine ddx_build_e(c_ddx, multipoles, mmax, phi_cav, e_cav, &
-        & nsph, ncav) bind(C)
-    type(c_ptr), intent(in), value :: c_ddx
-    type(ddx_setup), pointer :: ddx
-    integer(c_int), intent(in), value :: mmax, nsph, ncav
-    real(c_double), intent(in) :: multipoles((mmax + 1)**2, nsph)
-    real(c_double), intent(out) :: phi_cav(ncav), e_cav(3, ncav)
-    call c_f_pointer(c_ddx, ddx)
-    call build_e(ddx%params, ddx%constants, ddx%workspace, multipoles, mmax, &
-        & phi_cav, e_cav)
-end
-
-subroutine ddx_build_phi(c_ddx, multipoles, mmax, phi_cav, nsph, ncav) bind(C)
-    type(c_ptr), intent(in), value :: c_ddx
-    type(ddx_setup), pointer :: ddx
-    integer(c_int), intent(in), value :: mmax, nsph, ncav
-    real(c_double), intent(in) :: multipoles((mmax + 1)**2, nsph)
+    integer(c_int), intent(in), value :: nsph, ncav, nmultipoles
+    real(c_double), intent(in) :: multipoles(nmultipoles, nsph)
     real(c_double), intent(out) :: phi_cav(ncav)
+    integer :: mmax
     call c_f_pointer(c_ddx, ddx)
-    call build_phi(ddx%params, ddx%constants, ddx%workspace, multipoles, &
-        & mmax, phi_cav)
+    mmax = int(sqrt(dble(nmultipoles)) - 1d0)
+    call build_phi(ddx%params, ddx%constants, ddx%workspace, multipoles, mmax, &
+            & phi_cav)
 end
 
-subroutine ddx_build_psi(c_ddx, multipoles, mmax, psi, nsph, lmax) bind(C)
+subroutine ddx_multipole_electrostatics_1(c_ddx, nsph, ncav, nmultipoles, multipoles, &
+        & phi_cav, e_cav) bind(C)
     type(c_ptr), intent(in), value :: c_ddx
     type(ddx_setup), pointer :: ddx
-    integer(c_int), intent(in), value :: mmax, nsph, lmax
-    real(c_double), intent(in) :: multipoles((mmax + 1)**2, nsph)
-    real(c_double), intent(out) :: psi((lmax + 1)**2, nsph)
+    integer(c_int), intent(in), value :: nsph, ncav, nmultipoles
+    real(c_double), intent(in) :: multipoles(nmultipoles, nsph)
+    real(c_double), intent(out) :: phi_cav(ncav), e_cav(3, ncav)
+    integer :: mmax
     call c_f_pointer(c_ddx, ddx)
+    mmax = int(sqrt(dble(nmultipoles)) - 1d0)
+    call build_e(ddx%params, ddx%constants, ddx%workspace, multipoles, mmax, &
+            & phi_cav, e_cav)
+end
+
+subroutine ddx_multipole_electrostatics_2(c_ddx, nsph, ncav, nmultipoles, multipoles, &
+        & phi_cav, e_cav, g_cav) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx
+    type(ddx_setup), pointer :: ddx
+    integer(c_int), intent(in), value :: nsph, ncav, nmultipoles
+    real(c_double), intent(in) :: multipoles(nmultipoles, nsph)
+    real(c_double), intent(out) :: phi_cav(ncav), e_cav(3, ncav), g_cav(3, 3, ncav)
+    integer :: mmax
+    call c_f_pointer(c_ddx, ddx)
+    mmax = int(sqrt(dble(nmultipoles)) - 1d0)
+    call build_g(ddx%params, ddx%constants, ddx%workspace, multipoles, mmax, &
+            & phi_cav, e_cav, g_cav)
+end
+
+subroutine ddx_multipole_psi(c_ddx, nbasis, nsph, nmultipoles, multipoles, psi) bind(C)
+    type(c_ptr), intent(in), value :: c_ddx
+    type(ddx_setup), pointer :: ddx
+    integer(c_int), intent(in), value :: nmultipoles, nsph, nbasis
+    real(c_double), intent(in) :: multipoles(nmultipoles, nsph)
+    real(c_double), intent(out) :: psi(nbasis, nsph)
+    integer :: mmax
+    call c_f_pointer(c_ddx, ddx)
+    mmax = int(sqrt(dble(nmultipoles)) - 1d0)
     call build_psi(ddx%params, ddx%constants, ddx%workspace, multipoles, &
         & mmax, psi)
 end
 
-subroutine ddx_grad_phi(c_ddx, c_state, multipoles, mmax, forces, e_cav, &
-        & nsph, ncav) bind(C)
+subroutine ddx_multipole_forces(c_ddx, c_state, nsph, ncav, nmultipoles, multipoles, &
+                & e_cav, forces) bind(C)
     type(c_ptr), intent(in), value :: c_ddx, c_state
     type(ddx_setup), pointer :: ddx
     type(ddx_state_type), pointer :: state
-    integer(c_int), intent(in), value :: mmax, nsph, ncav
-    real(c_double), intent(in) :: multipoles((mmax + 1)**2, nsph), &
+    integer(c_int), intent(in), value :: nmultipoles, nsph, ncav
+    integer :: mmax
+    real(c_double), intent(in) :: multipoles(nmultipoles, nsph), &
         & e_cav(3, ncav)
     real(c_double), intent(out) :: forces(3, nsph)
     call c_f_pointer(c_ddx, ddx)
     call c_f_pointer(c_state, state)
+    mmax = int(sqrt(dble(nmultipoles)) - 1d0)
     call grad_phi(ddx%params, ddx%constants, ddx%workspace, state, mmax, &
         & multipoles, forces, e_cav)
 end
