@@ -335,8 +335,10 @@ subroutine contract_grad_C(params, constants, workspace, Xr, Xe, Xadj_r_sgrid, &
 
     call contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_sgrid, &
         & Xadj_e_sgrid, Xadj_r, Xadj_e, force, diff_re)
+    if (workspace % error_flag .eq. 1) return
     call contract_grad_C_worker1(params, constants, workspace, Xr, Xe, Xadj_r_sgrid, &
         & Xadj_e_sgrid, diff_re, force)
+    if (workspace % error_flag .eq. 1) return
 
 end subroutine contract_grad_C
 
@@ -364,8 +366,10 @@ subroutine contract_grad_f(params, constants, workspace, sol_adj, sol_sgrid, &
 
     call contract_grad_f_worker1(params, constants, workspace, sol_adj, sol_sgrid, &
         & gradpsi, force)
+    if (workspace % error_flag .eq. 1) return
     call contract_grad_f_worker2(params, constants, workspace, sol_sgrid, gradpsi, &
         & normal_hessian_cav, icav_g, force)
+    if (workspace % error_flag .eq. 1) return
 
 end subroutine contract_grad_f
 
@@ -667,8 +671,12 @@ subroutine contract_grad_C_worker1(params, constants, workspace, Xr, Xe, Xadj_r_
     allocate(phi_n_r(params % ngrid, params % nsph), &
         & phi_n_e(params % ngrid, params % nsph), &
         & diff_re_sgrid(params % ngrid, params % nsph), stat=istat)
-    if (istat.ne.0) stop 1
-
+    if (istat.ne.0) then
+        workspace % error_flag = 1
+        workspace % error_message = "allocation error in ddx_contract_grad_C_worker1"
+        return
+    end if
+!
     sum_int = zero
     sum_r = zero
     sum_e = zero
@@ -686,7 +694,11 @@ subroutine contract_grad_C_worker1(params, constants, workspace, Xr, Xe, Xadj_r_
     if (params % fmm .eq. 0) then
         allocate(coefY_d(constants % ncav, params % ngrid, params % nsph), &
             & stat=istat)
-        if (istat.ne.0) stop 1
+        if (istat.ne.0) then
+            workspace % error_flag = 1
+            workspace % error_message = "allocation error in fmm ddx_contract_grad_C_worker1"
+            return
+        end if
         coefY_d = zero
         ! Compute  summation over l0, m0
         ! Loop over the sphers j
@@ -859,10 +871,18 @@ subroutine contract_grad_C_worker1(params, constants, workspace, Xr, Xe, Xadj_r_
     end do
 
     deallocate(phi_n_r, phi_n_e, diff_re_sgrid, stat=istat)
-    if (istat.ne.0) stop 1
+    if (istat.ne.0) then
+        workspace % error_flag = 1
+        workspace % error_message = "deallocation error in ddx_contract_grad_C_worker1"
+        return
+    end if
     if (allocated(coefY_d)) then
         deallocate(coefY_d, stat=istat)
-        if (istat.ne.0) stop 1
+        if (istat.ne.0) then
+            workspace % error_flag = 1
+            workspace % error_message = "deallocation error in ddx_contract_grad_C_worker1"
+            return
+        end if
     end if
 end subroutine contract_grad_C_worker1
 
@@ -884,8 +904,8 @@ end subroutine contract_grad_C_worker1
 !! @param[inout] force     : Force
 !! @param[out] diff_re     : epsilon_1/epsilon_2 * l'/r_j[Xr]_jl'm'
 !!                         - (i'_l'(r_j)/i_l'(r_j))[Xe]_jl'm'
-subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_sgrid, &
-    & Xadj_e_sgrid, Xadj_r, Xadj_e, force, diff_re)
+subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, &
+        & Xadj_r_sgrid, Xadj_e_sgrid, Xadj_r, Xadj_e, force, diff_re)
     !! input/output
     type(ddx_params_type), intent(in) :: params
     type(ddx_constants_type), intent(in) :: constants
@@ -934,7 +954,11 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
         & diff1(constants % nbasis0, params % nsph), &
         & diff1_grad((constants % lmax0+2)**2, 3, params % nsph), &
         & l2l_grad((params % pl+2)**2, 3, params % nsph), stat=istat)
-    if (istat.ne.0) stop 1
+    if (istat.ne.0) then
+        workspace % error_flag = 1
+        workspace % error_message = "allocation error in ddx_contract_grad_C_worker2"
+        return
+    end if
 
     ! Setting initial values to zero
     SK_rijn = zero
@@ -942,6 +966,8 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
 
     diff_re = zero
     ! Compute l'/r_j[Xr]_jl'm' -(i'_l'(r_j)/i_l'(r_j))[Xe]_jl'm'
+    !$omp parallel do default(none) shared(params,diff_re,constants,xr,xe) &
+    !$omp private(jsph,l,m,ind) schedule(dynamic)
     do jsph = 1, params % nsph
       do l = 0, params % lmax
         do m = -l,l
@@ -954,6 +980,8 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
 
     ! diff0 = Pchi * diff_re, linear scaling
     diff0 = zero
+    !$omp parallel do default(none) shared(params,constants,diff_re,diff0, &
+    !$omp diff1,diff1_grad) private(jsph,l0,ind0) schedule(dynamic)
     do jsph = 1, params % nsph
       do l0 = 0, constants % lmax0
         do ind0 = l0*l0+1, l0*l0+2*l0+1
@@ -1026,6 +1054,8 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
             & params % lmax, workspace % tmp_sph, one, &
             & phi_in)
         ! Make phi_in zero at internal grid points
+        !$omp parallel do default(none) shared(params,constants,phi_in) &
+        !$omp private(isph,igrid) schedule(dynamic)
         do isph = 1, params % nsph
             do igrid = 1, params % ngrid
                 if (constants % ui(igrid, isph) .eq. zero) then
@@ -1034,6 +1064,8 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
             end do
         end do
         ! Get gradients of the L2L
+        !$omp parallel do default(none) shared(params,constants,workspace, &
+        !$omp l2l_grad) private(isph,igrid,inode) schedule(dynamic)
         do isph = 1, params % nsph
             inode = constants % snode(isph)
             workspace % tmp_sph_l(:, isph) = workspace % tmp_node_l(:, inode)
@@ -1061,6 +1093,8 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
         ! Properly load adjoint multipole harmonics into tmp_sph2 that holds
         ! harmonics of a degree up to lmax+1
         if(constants % lmax0+1 .lt. params % pm) then
+            !$omp parallel do default(none) shared(params,constants, &
+            !$omp workspace) private(isph,inode) schedule(dynamic)
             do isph = 1, params % nsph
                 inode = constants % snode(isph)
                 workspace % tmp_sph2(1:(constants % lmax0+2)**2, isph) = &
@@ -1069,6 +1103,8 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
             end do
         else
             indl = (params % pm+1)**2
+            !$omp parallel do default(none) shared(params,constants,indl, &
+            !$omp workspace) private(isph,inode) schedule(dynamic)
             do isph = 1, params % nsph
                 inode = constants % snode(isph)
                 workspace % tmp_sph2(1:indl, isph) = &
@@ -1158,8 +1194,9 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
             if(constants % ui(igrid, ksph) .gt. zero) then
                 icav = icav + 1
                 do ind = 1, constants % nbasis
-                    sum_dim3(:,ind,ksph) = sum_dim3(:,ind,ksph) + &
-                        & constants % coefvec(igrid, ind, ksph)*diff_ep_dim3(:,icav)
+                    sum_dim3(:,ind,ksph) = sum_dim3(:,ind,ksph) &
+                        & + diff_ep_dim3(:,icav)*constants % ui(igrid, ksph) &
+                        & *constants % vwgrid(ind, igrid)
                 end do
             end if
         end do
@@ -1208,8 +1245,9 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
                 if(constants % ui(igrid, isph) .gt. zero) then
                   icav = icav + 1
                   do ind = 1, constants % nbasis
-                    sum_dim3(:,ind,isph) = sum_dim3(:,ind,isph) + &
-                        & constants % coefvec(igrid, ind, isph)*diff_ep_dim3(:,icav)
+                    sum_dim3(:,ind,isph) = sum_dim3(:,ind,isph) &
+                        & + diff_ep_dim3(:,icav)*constants % ui(igrid, isph) &
+                        & *constants % vwgrid(ind, igrid)
                   end do
                 end if
               end do
@@ -1230,7 +1268,11 @@ subroutine contract_grad_C_worker2(params, constants, workspace, Xr, Xe, Xadj_r_
     end do
     deallocate(diff_ep_dim3, phi_in, sum_dim3, diff1_grad, l2l_grad, &
         & diff0, diff1, stat=istat)
-    if (istat.ne.0) stop 1
+    if (istat.ne.0) then
+        workspace % error_flag = 1
+        workspace % error_message = "deallocation error in ddx_contract_grad_C_worker2"
+        return
+    end if
 end subroutine contract_grad_C_worker2
 
 !> Subroutine to compute the derivative of U_i(x_in) and \bf{k}^j_l0(x_in)Y^j_l0m0(x_in)
@@ -1301,7 +1343,11 @@ subroutine contract_grad_f_worker1(params, constants, workspace, sol_adj, sol_sg
         & sum_Sjin(params % ngrid, params % nsph), &
         & l2l_grad((params % pl+2)**2, 3, params % nsph), &
         & diff0(constants % nbasis0, params % nsph), stat=istat)
-    if (istat.ne.0) stop 1
+    if (istat.ne.0) then
+        workspace % error_flag = 1
+        workspace % error_message = "allocation error in ddx_grad_f_worker1"
+        return
+    end if
 
 
     ! Setting initial values to zero
@@ -1510,9 +1556,10 @@ subroutine contract_grad_f_worker1(params, constants, workspace, sol_adj, sol_sg
             if(constants % ui(igrid, ksph) .gt. zero) then
                 icav = icav + 1
                 do ind = 1, constants % nbasis
-                    sum_dim3(:,ind,ksph) = sum_dim3(:,ind,ksph) + &
-                        & -(params % epsp/params % eps)* &
-                        & constants % coefvec(igrid, ind, ksph)*diff_ep_dim3(:,icav)
+                    sum_dim3(:,ind,ksph) = sum_dim3(:,ind,ksph) &
+                        & - (params % epsp/params % eps) &
+                        & *diff_ep_dim3(:,icav)*constants % ui(igrid, ksph) &
+                        & *constants % vwgrid(ind, igrid)
                 end do
             end if
         end do
@@ -1556,9 +1603,10 @@ subroutine contract_grad_f_worker1(params, constants, workspace, sol_adj, sol_sg
                 if(constants % ui(igrid, isph) .gt. zero) then
                   icav = icav + 1
                   do ind = 1, constants % nbasis
-                    sum_dim3(:,ind,isph) = sum_dim3(:,ind,isph) + &
-                                          & -(params % epsp/params % eps)* &
-                                          & constants % coefvec(igrid, ind, isph)*diff_ep_dim3(:,icav)
+                    sum_dim3(:,ind,isph) = sum_dim3(:,ind,isph) &
+                      & - (params % epsp/params % eps) &
+                      & *diff_ep_dim3(:,icav)*constants % ui(igrid, isph) &
+                      & *constants % vwgrid(ind, igrid)
                   end do
                 end if
               end do
@@ -1579,7 +1627,11 @@ subroutine contract_grad_f_worker1(params, constants, workspace, sol_adj, sol_sg
 
     deallocate(phi_in, diff_ep_dim3, sum_dim3, c0_d, c0_d1, &
           & c0_d1_grad, sum_Sjin, l2l_grad, diff0, stat=istat)
-    if (istat.ne.0) stop 1
+    if (istat.ne.0) then
+        workspace % error_flag = 1
+        workspace % error_message = "deallocation error in ddx_grad_f_worker1"
+        return
+    end if
 
 end subroutine contract_grad_f_worker1
 
@@ -1640,7 +1692,11 @@ subroutine contract_grad_f_worker2(params, constants, workspace, sol_sgrid, grad
     allocate(phi_n(params % ngrid, params % nsph), &
         & phi_n2(params % ngrid, params % nsph), &
         & gradpsi_grid(params % ngrid, params % nsph), stat=istat)
-    if (istat.ne.0) stop 1
+    if (istat.ne.0) then
+        workspace % error_flag = 1
+        workspace % error_message = "allocation error in ddx_grad_f_worker2"
+        return
+    end if
 
     ! Intial allocation of vectors
     sum_int = zero
@@ -1657,7 +1713,11 @@ subroutine contract_grad_f_worker2(params, constants, workspace, sol_sgrid, grad
     if (params % fmm .eq. 0) then
         allocate(coefY_d(constants % ncav, params % ngrid, params % nsph), &
             & stat=istat)
-        if (istat.ne.0) stop 1
+        if (istat.ne.0) then
+            workspace % error_flag = 1
+            workspace % error_message = "allocation error in fmm ddx_grad_f_worker2"
+            return
+        end if
         coefY_d = zero
         ! Compute  summation over l0, m0
         ! Loop over the sphers j
@@ -1791,10 +1851,18 @@ subroutine contract_grad_f_worker2(params, constants, workspace, sol_sgrid, grad
     end do
 
     deallocate(phi_n, phi_n2, gradpsi_grid, stat=istat)
-    if (istat.ne.0) stop 1
+    if (istat.ne.0) then
+        workspace % error_flag = 1
+        workspace % error_message = "deallocation error in ddx_grad_f_worker2"
+        return
+    end if
     if (allocated(coefY_d)) then
         deallocate(coefY_d, stat=istat)
-        if (istat.ne.0) stop 1
+        if (istat.ne.0) then
+            workspace % error_flag = 1
+            workspace % error_message = "deallocation error in ddx_grad_f_worker2"
+            return
+        end if
     end if
 end subroutine contract_grad_f_worker2
 
